@@ -12,7 +12,7 @@
 //! # Basic example
 //!
 //! ```
-//! use bakery::load_struct_from_string;
+//! use bakery::load_from_string;
 //! use bakery_derive::Recipe;
 //! use serde::Deserialize;
 //!
@@ -23,7 +23,7 @@
 //!     fullscreen: bool
 //! }
 //!
-//! let config: GameConfig = load_struct_from_string("width: 1024, height: 768, fullscreen: true");
+//! let config: GameConfig = load_from_string("width: 1024, height: 768, fullscreen: true");
 //! ```
 
 use num_bigint::{BigInt, Sign};
@@ -187,8 +187,9 @@ pub enum NodeContent {
     RecTypeInst {
         tid: RecTypeId,
     },
-    /// Generic type node which may be child of RecStruct or native generic types such as RecList
-    /// `index` field correspond to the index of the generic type in the parent type.
+    /// Generic type node which may be child of `RecStruct` or native generic types such as
+    /// `RecList`.
+    /// `index` field corresponds to the index of the generic type in the parent type.
     RecGeneric {
         index: u32,
     },
@@ -309,6 +310,11 @@ impl NodeTree {
         nid
     }
 
+    /// Create a recipe structure node with no name and no parent
+    pub fn create_root_struct(&mut self) -> u32 {
+        self.create(Node::new_anonymous(NodeContent::RecStruct))
+    }
+
     /// Create a recipe structure node and return node Id.
     ///
     /// # Arguments
@@ -346,14 +352,99 @@ impl NodeTree {
         nid
     }
 
+    /// Create a recipe enumeration node and return node Id.
+    ///
+    /// # Arguments
+    ///
+    /// * `parent` - Parent node
+    /// * `name` - Enumeration name
+    pub fn create_enum(&mut self, parent: Option<u32>, name: &str, nid_key_type: u32) -> u32 {
+        self.create_with_parent(
+            parent,
+            Node {
+                name: Some(name.to_string()),
+                source: None,
+                content: NodeContent::RecEnum {
+                    key_type: RecTypeId::Id(nid_key_type)
+                }
+            }
+        )
+    }
+
+    /// Create an enumeration member node and return node Id.
+    ///
+    /// # Arguments
+    ///
+    /// * `parent` - Parent node
+    /// * `name` - Enumeration member name
+    pub fn create_enum_member(&mut self, parent: u32, name: &str, value: BigInt) -> u32 {
+        self.create_with_parent(
+            Some(parent),
+            Node {
+                name:Some(name.to_string()),
+                source: None,
+                content: NodeContent::RecEnumItem { value: value }
+            }
+        )
+    }
+
+    /// Create a tuple node and return node Id.
+    ///
+    /// # Arguments
+    ///
+    /// * `parent` - Parent node
+    pub fn create_tuple(&mut self, parent: Option<u32>) -> u32 {
+        self.create_with_parent(
+            parent,
+            Node::new_anonymous( NodeContent::RecTuple )
+        )
+    }
+
+    /// Create a tuple item node and return node Id.
+    ///
+    /// # Arguments
+    ///
+    /// * `parent` - Parent node
+    /// * `ty` - Type node
+    pub fn create_tuple_member(&mut self, parent: u32, ty: u32) -> u32 {
+        self.create_with_parent(
+            Some(parent),
+            Node::new_anonymous(
+                NodeContent::RecTupleMember {
+                    tid: RecTypeId::Id(ty)
+                }
+            )
+        )
+    }
+
+    /// Parse a type recipe from a string, return recipe node Id.
+    ///
+    /// # Arguments
+    ///
+    /// * `rec` - Recipe string
+    pub fn parse_recipe_string(&mut self, rec: &str) -> Result<u32, ParseError> {
+        let rec = rec.trim_start().trim_end();
+        let mut pairs = MyParser::parse(Rule::rec_type_anonymous, rec).unwrap();
+        let pair = pairs.next().unwrap();
+        let span = pair.as_span();
+        assert!(pairs.next().is_none());
+        // Check the the whole string has been parsed.
+        // pest will match as much as it can, so if there's garbage at the end, it will be ignored.
+        // We don't want to ignore this garbage.
+        if span.end() != rec.len() {
+            return Err(ParseError::IncompleteRecParse(span.end()));
+        }
+        Ok(self.parse_rec_type(Rc::new(rec.to_string()), pair))
+    }
+
     /// Build recipe from a string, return recipe node Id.
     ///
     /// # Arguments
     ///
     /// * `rec` - Recipe string
-    fn parse_recipe_string(&mut self, rec: &str) -> Result<u32, ParseError> {
+    pub fn parse_struct_recipe_string(&mut self, rec: &str) -> Result<u32, ParseError> {
         // Parse recipe
-        let rec = rec.trim_end();
+        let rec = rec.trim_start().trim_end();
         let mut pairs = MyParser::parse(Rule::file_rec, rec).unwrap();
         let pair = pairs.next().unwrap();
         let span = pair.as_span();
@@ -381,7 +472,7 @@ impl NodeTree {
     ///
     /// * `dat` - Data string
     pub fn parse_dat_value_string(&mut self, dat: &str) -> Result<u32, ParseError> {
-        let dat = dat.trim_end();
+        let dat = dat.trim_start().trim_end();
         let mut pairs = MyParser::parse(Rule::dat_value, dat).unwrap();
         let pair = pairs.next().unwrap();
         let span = pair.as_span();
@@ -401,7 +492,7 @@ impl NodeTree {
     ///
     /// * `dat` - Data string, struct format without the braces.
     pub fn parse_dat_map_string(&mut self, dat: &str) -> Result<u32, ParseError> {
-        let dat = dat.trim_end();
+        let dat = dat.trim_start().trim_end();
         let mut pairs = MyParser::parse(Rule::file_dat, dat).unwrap();
         let pair = pairs.next().unwrap();
         let span = pair.as_span();
@@ -1011,7 +1102,7 @@ fn write_int_check_bounds(
     }
 }
 
-pub struct Compiler<'a> {
+struct Compiler<'a> {
     tree: NodeTree,
     io: &'a mut dyn std::io::Write,
     errors: Vec<CompilationError>,
@@ -1612,19 +1703,24 @@ impl Compiler<'_> {
 
     pub fn compile(&mut self, rec: &str, dat: &str, print_errs: bool) -> Result<(), ()> {
         self.tree.clear();
+        let node_root = self.tree.create_root_struct();
+        self.tree.populate_natives(node_root); // Add standard types
         match self.tree.parse_recipe_string(rec) {
-            Ok(main_struct_node) => {
-                // Add standard types
-                self.tree.populate_natives(main_struct_node);
-                self.resolve_types(main_struct_node);
+            Ok(node_rec) => {
+                self.tree.child(node_root, node_rec);
+                self.resolve_types(node_root);
                 // resolve_types may generate errors if some types are unknown. In that case, the tree
                 // will have dangling references to types, which is not supported for the rest of the
                 // compilation.
                 if self.errors.len() == 0 {
                     // Parse data
-                    match self.tree.parse_dat_map_string(dat) {
-                        Ok(data_node) => {
-                            if let Err(e) = self.write(main_struct_node, data_node) {
+                    let parsed_dat = match self.tree.get(node_rec).content {
+                        NodeContent::RecStruct => self.tree.parse_dat_map_string(dat),
+                        _ => self.tree.parse_dat_value_string(dat)
+                    };
+                    match parsed_dat {
+                        Ok(node_dat) => {
+                            if let Err(e) = self.write(node_rec, node_dat) {
                                 self.error(CompilationError::IOError(e));
                             }
                         }
@@ -1752,6 +1848,7 @@ fn print_errors(tree: &NodeTree, errors: &Vec<CompilationError>) {
     }
 }
 
+#[derive(Debug)]
 pub enum LoadError {
     CompilationErrors,
     InvalidRecExtension,
@@ -1794,6 +1891,25 @@ fn is_compilation_required(
     }
 }
 
+/// Write the binary representation of string data to be compiled, with the recipe given as a
+/// string.
+///
+/// # Arguments
+///
+/// * `dest` - A writable stream
+/// * `rec` - Recipe string
+/// * `dat` - Data string
+pub fn write_from_string_with_recipe<'a>(out: &'a mut dyn std::io::Write, rec: &str, dat: &str)
+    -> Result<(), LoadError>
+{
+    let mut compiler = Compiler::new(out);
+    if let Err(()) = compiler.compile(rec, dat, true) {
+        return Err(LoadError::CompilationErrors)
+    } else {
+        Ok(())
+    }
+}
+
 /// Load an object from a data file, with recipe built using Recipe trait.
 ///
 /// If the binary image of the data file does not exist or is deprecated, it is built and cached.
@@ -1829,7 +1945,7 @@ where
     return Ok(bincode::deserialize_from(file).unwrap());
 }
 
-/// Load an object from a string, with recipe built using `Recipe` trait.
+/// Load data from a string, with recipe built using [`Recipe`] trait.
 ///
 /// # Arguments
 ///
@@ -1837,11 +1953,28 @@ where
 ///
 /// # Example
 ///
+/// This example shows how to load a structure from a string:
 /// ```
 /// use bakery::load_from_string;
 /// use bakery_derive::Recipe;
 /// use serde::Deserialize;
 ///
+/// #[derive(Recipe, Deserialize, Debug, PartialEq)]
+/// struct GameConfig {
+///     width: u32,
+///     height: u32,
+///     fullscreen: bool
+/// }
+///
+/// let config: GameConfig = load_from_string("width: 1024, height: 768, fullscreen: true");
+/// assert_eq!(config, GameConfig { width: 1024, height: 768, fullscreen: true });
+/// ```
+///
+/// This example shows how to load a list from a string. Note that the [`Recipe`] trait for [`Vec`] is
+/// implemented by the library.
+/// ```
+/// use bakery::load_from_string;
+///
 /// let values: Vec<i32> = load_from_string("[1, 2, 3]");
 /// assert_eq!(values, vec![1, 2, 3]);
 /// ```
@@ -1852,44 +1985,10 @@ where
     let mut bin = Vec::<u8>::new();
     let mut compiler = Compiler::new(&mut bin);
     let nid_rec = T::recipe(&mut compiler.tree);
-    let nid_dat = compiler.tree.parse_dat_value_string(dat).unwrap();
-    compiler.resolve_types(nid_rec);
-    compiler.write(nid_rec, nid_dat).unwrap();
-    bincode::deserialize_from(&bin[..]).unwrap()
-}
-
-/// Load a structure from a string, with recipe built using `Recipe` trait, assuming the data
-/// string is a list of assignments (structure without enclosing braces).
-///
-/// # Arguments
-///
-/// * `dat` - Data string
-///
-/// # Example
-///
-/// ```
-/// use bakery::load_struct_from_string;
-/// use bakery_derive::Recipe;
-/// use serde::Deserialize;
-///
-/// #[derive(Recipe, Deserialize, Debug, PartialEq)]
-/// struct GameConfig {
-///     width: u32,
-///     height: u32,
-///     fullscreen: bool
-/// }
-///
-/// let config: GameConfig = load_struct_from_string("width: 1024, height: 768, fullscreen: true");
-/// assert_eq!(config, GameConfig { width: 1024, height: 768, fullscreen: true });
-/// ```
-pub fn load_struct_from_string<T>(dat: &str) -> T
-where
-    T: Recipe + DeserializeOwned,
-{
-    let mut bin = Vec::<u8>::new();
-    let mut compiler = Compiler::new(&mut bin);
-    let nid_rec = T::recipe(&mut compiler.tree);
-    let nid_dat = compiler.tree.parse_dat_map_string(dat).unwrap();
+    let nid_dat = match compiler.tree.get(nid_rec).content {
+        NodeContent::RecStruct => compiler.tree.parse_dat_map_string(dat).unwrap(),
+        _ => compiler.tree.parse_dat_value_string(dat).unwrap()
+    };
     compiler.resolve_types(nid_rec);
     compiler.write(nid_rec, nid_dat).unwrap();
     bincode::deserialize_from(&bin[..]).unwrap()
